# Copyright (c) 2009 Mathias Kaerlev.
# See LICENSE for details.

# Copyright (c) 2009 Mathias Kaerlev.
# See LICENSE for details.

import time

from twisted.internet import reactor
from twisted.internet.protocol import ClientFactory

from pySndo.network.protocol import GameProtocol, GameFactory, \
    InfoProtocol, InfoFactory

# info things
    
from pySndo.network.info import GetInitial, RequestMapLine, GetMapSending, \
    InitialMessage, Error, MapLine, MapSending, MapReceiver

# game things

from pySndo.network.game import Message

from pySndo.network.packetloaders.welcome import Welcome, SetName
from pySndo.network.packetloaders.playerlist import PlayerList
from pySndo.network.packetloaders.map import CurrentMap, SetMap, MapStart, \
    BombedArea, SpawnItem, GotMedipack
from pySndo.network.packetloaders.chat import Chat, TeamChat, Typing, \
    DoneTyping, ServerMessage
from pySndo.network.packetloaders.ping import Ping, Pong
from pySndo.network.packetloaders.join import JoinGame, RequestSpawn, \
    SpawnGranted
from pySndo.network.packetloaders.team import SetTeam, TeamScore, \
    IncrementScore, SelfKill
from pySndo.network.packetloaders.movement import Moving, Stopped, \
    ChangeDirection, LookDirection, Spawn
from pySndo.network.packetloaders.battle import Reload, Shoot, Hit, HeavyHit, \
    Killed, UseItem, ThrowFlashbang, ThrowSmoke, ThrowMolotov, \
    itemToId, weapons, dieWeapons
from pySndo.network.packetloaders.player import PlayerJoined, PlayerLeft, \
    PlayerKicked, GotBanned, PlayerKills, SkinChange
from pySndo.network.packetloaders.status import Closing, Restart, ServerFull, \
    TimeLeft, NextMap, ServerStatus

# packetloader bases

from pySndo.network.packetloaders.common import PrefixSend

# logic

from pySndo.game.player import Player

from pySndo.game.crate import guns, grenades

class OtherClient(object):
    """
    This class represents another client.
    
    @ivar name: Name of the client.
    @ivar id: ID of the cllient.
    @ivar team: Team of the client
    @type team: L{Police} or L{Terrorist} instance.
    @type player: L{pySndo.game.player.Player} instance.
    """
    name = None
    id = None
    latency = None
    team = None
    player = None

    def __init__(self):
        self.player = Player()

class ClientGameProtocol(GameProtocol):
    _isClient = True
    player = None
    
    _spawnX = None
    _spawnY = None
    
    _requestTime = None
    
    # implementation-specifics
    
    def packetReceived(self, packet):
        packetType = packet.type

        if isinstance(packetType, Welcome):
            self.id = packetType.id
            self.sendType(JoinGame())
            self.isAccepted = True
            self.player = Player()
            self.loggedIn()
            
        elif isinstance(packetType, Closing):
            self.serverClosed()
            
        elif isinstance(packetType, GotBanned):
            if packetType.playerId == self.id:
                self.gotBanned()
            else:
                newClient = self.factory.playerFromId(packetType.playerId)
                self.playerBanned(newClient)
                
        elif isinstance(packetType, PlayerKicked):
            if packetType.playerId == self.id:
                self.gotKicked()
            else:
                newClient = self.factory.playerFromId(packetType.playerId)
                self.playerKicked(newClient)
            
        elif isinstance(packetType, ServerFull):
            self.serverFull()
            
        elif isinstance(packetType, Restart):
            self.factory.police.score = 0
            self.factory.terrorist.score = 0
            self.factory.bombedAreas = []
            self.sessionRestarted()
            
        elif isinstance(packetType, Ping):
            self.sendType(Pong())

        elif isinstance(packetType, TeamScore):
            self.factory.police.score = packetType.policePoints
            self.factory.terrorist.score = packetType.terroristPoints
            self.teamScoresReceived(packetType.terroristPoints, packetType.policePoints)

        elif isinstance(packetType, CurrentMap):
            self.factory.map = self.factory.loadMap(packetType.value)
            self.mapChanged(self.factory.map)
            
        elif isinstance(packetType, SetMap):
            self.factory.map = self.factory.loadMap(packetType.mapName)
            for connection in self.factory.connections:
                connection.player.alive = False
                connection.player.setTeam('Generic')
            self.mapChanged(self.factory.map, crc = packetType.mapCrc)
            
        elif isinstance(packetType, PlayerList):
            newClient = OtherClient()
            newClient.name = packetType.name
            newClient.id = packetType.id
            newClient.team = self.factory.teamFromName(packetType.getTeam())
            newClient.player.alive = packetType.alive
            newClient.player.kills = packetType.kills
            newClient.player.deaths = packetType.deaths
            newClient.player.typing = packetType.typing
            newClient.latency = packetType.latency
            newClient.player.movement.set(packetType.x, packetType.y, 
                                          packetType.direction)
            if packetType.moving:
                newClient.player.movement.start()
            self.factory.connections.append(newClient)
            self.playerExists(newClient)
            
        elif isinstance(packetType, PlayerJoined):
            newClient = OtherClient()
            newClient.name = packetType.name
            newClient.id = packetType.id
            self.factory.connections.append(newClient)
            self.playerJoined(newClient)
            
        elif isinstance(packetType, PlayerLeft):
            newClient = self.factory.playerFromId(packetType.id)
            self.factory.connections.remove(newClient)
            self.playerLeft(newClient)
            
        elif isinstance(packetType, SetTeam):
            connection = self.factory.playerFromId(packetType.id)
            team = self.factory.teamFromName(packetType.getTeam())
            if connection.player.team:
                self.playerChangedTeam(connection, team)
            else:
                self.playerJoinedTeam(connection, team)
            connection.player.team = team
            
        elif isinstance(packetType, Typing):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.typing = True
            
        elif isinstance(packetType, ServerMessage):
            self.messageReceived(packetType.value)
            
        elif isinstance(packetType, ServerStatus):
            self.statusReceived(packetType.value)
            
        elif isinstance(packetType, Chat):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.typing = False
            self.playerMessageReceived(connection, packetType.value)
            
        elif isinstance(packetType, TeamChat):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.typing = False
            self.teamMessageReceived(connection, packetType.value)

        elif isinstance(packetType, DoneTyping):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.typing = False
            
        elif isinstance(packetType, Moving):
            try:
                connection = self.factory.playerFromId(packetType.id)
            except IndexError:
                # player couldn't be found, must have left
                return
            connection = self.factory.playerFromId(packetType.id)
            movement = connection.player.movement
            x, y = movement.getPosition()
            movement.set(x, y, packetType.direction)
            movement.start()
            self.playerMoving(connection)
            
        elif isinstance(packetType, ChangeDirection):
            try:
                connection = self.factory.playerFromId(packetType.id)
            except IndexError:
                # player couldn't be found, must have left
                return
            connection = self.factory.playerFromId(packetType.id)
            movement = connection.player.movement
            movement.set(packetType.x, packetType.y, packetType.direction)
            movement.start()
            self.playerChangedDirection(connection)
            
        elif isinstance(packetType, Stopped):
            try:
                connection = self.factory.playerFromId(packetType.id)
            except IndexError:
                # player couldn't be found, must have left
                return
            movement = connection.player.movement
            movement.stop()
            movement.set(packetType.x, packetType.y, movement.direction)
            self.playerStopped(connection)
            
        elif isinstance(packetType, LookDirection):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.look_direction = packetType.direction
            self.playerLooked(connection)
            
        elif isinstance(packetType, Reload):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.reloading = packetType.reloading
            
        elif isinstance(packetType, Shoot):
            connection = self.factory.playerFromId(packetType.id)
            self.playerShot(connection, packetType.getWeapon(), packetType.showOnly,
                packetType.hitX, packetType.hitY)
                
        elif isinstance(packetType, Hit):
            if packetType.targetPlayer == self.id:
                sourcePlayer = self.factory.playerFromId(packetType.id)
                self.gotHit(sourcePlayer, packetType.getWeapon(), 
                    packetType.headshot, packetType.bloodAmount, packetType.damage)
            else:
                sourcePlayer = self.factory.playerFromId(packetType.id)
                targetPlayer = self.factory.playerFromId(packetType.targetPlayer)
                self.playerHit(sourcePlayer, targetPlayer, packetType.getWeapon(), 
                    packetType.headshot, packetType.bloodAmount, packetType.damage)
                
        elif isinstance(packetType, HeavyHit):
            if packetType.targetPlayer == self.id:
                sourcePlayer = self.factory.playerFromId(packetType.id)
                self.gotHit(sourcePlayer, packetType.getItem(), 
                    packetType.headshot, packetType.bloodAmount, packetType.damage)
            else:
                sourcePlayer = self.factory.playerFromId(packetType.id)
                targetPlayer = self.factory.playerFromId(packetType.targetPlayer)
                self.playerHit(sourcePlayer, targetPlayer, packetType.getItem(), 
                    packetType.headshot, packetType.bloodAmount, packetType.damage)

        elif isinstance(packetType, UseItem):
            connection = self.factory.playerFromId(packetType.id)
            self.playerUsedItem(connection, packetType.getItem(), 
                packetType.direction, packetType.power)
                
        elif packetType.__class__ in (ThrowFlashbang, ThrowMolotov, 
                                      ThrowSmoke):
            connection = self.factory.playerFromId(packetType.id)
            self.playerUsedItem(connection, packetType.getWeapon(), 
                packetType.direction, packetType.power)
                
        elif isinstance(packetType, MapStart):
            for connection in self.factory.connections:
                connection.player.health = 100
                connection.player.alive = True
            self.roundStarted()
            
        elif isinstance(packetType, IncrementScore):
            winningTeam = self.factory.teamFromName(packetType.getTeam())
            self.roundEnded(winningTeam)
        
        elif isinstance(packetType, Spawn):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.alive = True
            if packetType.latency:
                connection.latency = packetType.latency
            movement = connection.player.movement
            movement.stop()
            movement.set(packetType.x, packetType.y, movement.direction)
            self.playerSpawned(connection, packetType.x, packetType.y)
            
        elif isinstance(packetType, SpawnGranted):
            self._spawnGranted()
            
        elif isinstance(packetType, SelfKill):
            # ignore for now, no need for this
            pass
            
        elif isinstance(packetType, Killed):
            if packetType.killedPlayer == self.id:
                killedPlayer = self
                killedPlayer.player.alive = False
                killedPlayer.player.deaths += 1
                killingPlayer = self.factory.playerFromId(packetType.killingPlayer)
                self.gotKilled(killingPlayer, packetType.getDieType())
            else:
                killedPlayer = self.factory.playerFromId(packetType.killedPlayer)
                killedPlayer.player.alive = False
                killedPlayer.player.deaths += 1
                killingPlayer = self.factory.playerFromId(packetType.killingPlayer)
                self.playerKilled(killingPlayer, killedPlayer, packetType.getDieType())
            
        elif isinstance(packetType, PlayerKills):
            connection = self.factory.playerFromId(packetType.id)
            connection.player.kills = packetType.kills
            
        elif isinstance(packetType, SpawnItem):
            newCrate = packetType.getCrateType()(self.factory.crates)
            self.factory.crates.append(newCrate)
            newCrate.x = packetType.x
            newCrate.y = packetType.y
            for key, value in self.factory.initial.weapons.iteritems():
                if value == False and key in newCrate.possibleContent:
                    newCrate.possibleContent.remove(key)
            newCrate.removeCrate().addCallback(self.crateDestroyed)
            self.crateSpawned(newCrate)
            
        elif isinstance(packetType, BombedArea):
            self.factory.bombedAreas.append(packetType)
            self.bombedArea(packetType.x, packetType.y)
            
        else:
            raise ValueError('unknown packet: %s %s %s %s' % (packet.way, packet.id, 
                packetType.__class__.__name__, vars(packetType)))
        
    def sendType(self, packetType):
        newMessage = Message()
        newMessage.type = packetType
        if isinstance(packetType, PrefixSend) and not packetType.id:
            packetType.id = self.id # set the id to ourselves
        self.sendLine(newMessage.generate())
    
    # overridables

    def loggedIn(self):
        """
        Called upon logging in.
        """
        raise NotImplementedError
        
    def teamScoresReceived(self, terrorist, police):
        """
        Called when team scores have been received.
        """
        raise NotImplementedError
        
    def serverClosed(self):
        """
        Called when the server has closed intentionally.
        """
        raise NotImplementedError
        
    def serverFull(self):
        """
        Called upon joining if the server is full.
        """
        raise NotImplementedError
        
    def sessionRestarted(self):
        """
        Called upon session restart.
        """
        raise NotImplementedError
        
    def gotKicked(self):
        """
        Called upon getting kicked from the server.
        """
        raise NotImplementedError
        
    def gotBanned(self):
        """
        Called upon getting banned from the server.
        """
        raise NotImplementedError
        
    def playerJoined(self, player):
        """
        Called when a player has joined the server.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def playerExists(self, player):
        """
        Called for every player already on the server
        when joining.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError

    def playerLeft(self, player):
        """
        Called when a player has left the server.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def playerKicked(self, player):
        """
        Called when a player has been kicked from the server.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def playerBanned(self, player):
        """
        Called when a player has been banned from the server.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def messageReceived(self, message):
        """
        Called upon receiving a 'noticable' message from the server.
        """
        raise NotImplementedError
        
    def statusReceived(self, message):
        """
        Called upon receiving a server status message from the
        server.
        """
        raise NotImplementedError
        
    def playerMessageReceived(self, player, message):
        """
        Called when a player has sent a chat message.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def teamMessageReceived(self, player, message):
        """
        Called when a player has sent a team message.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def playerJoinedTeam(self, player, team):
        """
        Called when a player has joined a team.
        
        @type player: L{OtherClient} instance
        
        @type team: L{Police} or L{Terrorist} instance.
        """
        raise NotImplementedError
        
    def playerChangedTeam(self, player, team):
        """
        Called when a player has changed to a new team.
        
        @type player: L{OtherClient} instance
        
        @type team: L{Police} or L{Terrorist} instance.
        """
        raise NotImplementedError
        
    def playerMoving(self, player):
        """
        Called when a player has started movement.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
        
    def playerChangedDirection(self, player):
        """
        Called when a player changed direction while moving.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def playerLooked(self, player):
        """
        Called when a player has changed look direction.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def playerStopped(self, player):
        """
        Called when a player has stopped movement.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def playerUsedItem(self, player, item, direction, power):
        """
        Called when a player has used a special item.
        
        @type player: L{OtherClient} instance
        
        @type item: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param direction: What direction the item was thrown at
        
        @param power: The strength the item was thrown with
        """
        raise NotImplementedError
        
    def playerSpawned(self, player, x, y):
        """
        Called when a player has spawned at round start.
        
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError
        
    def spawnGranted(self):
        """
        Called upon getting granted a spawn.
        """
        raise NotImplementedError
        
    def playerShot(self, player, weapon, showOnly, hitX, hitY):
        """
        Called when a player has missed a shot
        
        @param player: The shooting player
        @type player: L{OtherClient} instance
        
        @type weapon: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param showOnly: If False, hitX and hitY are valid
        
        @param hitX: The X coordinate where the bullet has collided
        @param hitY: The Y coordinate where the bullet has collided
        """
        raise NotImplementedError
        
    def playerHit(self, sourcePlayer, targetPlayer, weapon, headshot, bloodAmount, damage):
        """
        Called when a player has been hit by another player (either by gun
        or grenade)
        
        @param sourcePlayer: The shooting player
        @type sourcePlayer: L{OtherClient} instance
        @param targetPlayer: The hitted player
        @type targetPlayer: L{OtherClient} instance
        
        @type weapon: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param damage: Damage the target player has taken
        
        @param headshot: True if player hit in headshot mode
        
        @param bloodAmount: Amount of blood in range of 0-2 if not headshot,
            0-6 if headshot
        """
        raise NotImplementedError
        
    def playerKilled(self, killingPlayer, killedPlayer, dieType):
        """
        Called when a player has been killed by another player.
        
        @param killingPlayer: The killing player
        @type killingPlayer: L{OtherClient} instance
        @param killedPlayer: The player that was killed
        @type killedPlayer: L{OtherClient} instance
        @type dieType: See L{pySndo.network.packetloaders.battle.dieTypes}
        """
        raise NotImplementedError
        
    def gotHit(self, player, weapon, headshot, bloodAmount, damage):
        """
        Called when this client has been hit by another player.
        
        @param player: The shooting player
        @type player: L{OtherClient} instance
        
        @type weapon: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param damage: Damage taken
        
        @param headshot: True if player hit in headshot mode
        
        @param bloodAmount: Amount of blood in range of 0-2 if not headshot,
            0-6 if headshot
        """
        raise NotImplementedError
        
    def gotKilled(self, player, dieType):
        """
        Called when this client has been killed by another player.
        
        @param player: The killing player
        @type player: L{OtherClient} instance
        """
        raise NotImplementedError

    def mapChanged(self, newMap, crc = None):
        """
        Called upon map change.
        
        @type newMap: L{Map}
        @param crc: If specified, CRC of the new map from the server,
            make sure to check the automatically loaded map
            with the requested map CRC (they might differ).
        @type crc: int
        """
        raise NotImplementedError
        
    def bombedArea(self, x, y):
        """
        Called upon receiving a bombed area
        (area bombed with C4 previously)
        
        @param x: X coordinate of the bombed area
        @param y: Y coordinate of the bombed area
        """
        raise NotImplementedError

    def roundEnded(self, winningTeam):
        """
        (LTS mode only)
        Called upon the end of the round
        
        @param winningTeam: The winner of this round
        @type winningTeam: L{Terrorist} or L{Police}
        """
        raise NotImplementedError

    def roundStarted(self):
        """
        (LTS mode only)
        Called upon round start
        """
        raise NotImplementedError
        
    # actions
    
    def login(self, name):
        """
        This method is used to log into the server
        with the given name.
        """
        self.name = name
        newSet = SetName()
        newSet.value = name
        self.sendType(newSet)
        
    def joinTeam(self, team):
        teamObject = self.factory.teamFromName(team)
        newSet = SetTeam()
        newSet.setTeam(team)
        self.sendType(newSet)
        if self.player.team:
            self.sendType(SelfKill())
        self.player.team = teamObject
    
    def shoot(self, weapon, showOnly = False, hitX = None, hitY = None):
        newShoot = Shoot()
        newShoot.setWeapon(weapon)
        newShoot.hitX = hitX
        newShoot.hitY = hitY
        newShoot.direction = self.player.direction
        self.sendType(newShoot)
        
    def useItem(self, item, power):
        if item == MolotovCocktail:
            newUse = ThrowMolotov()
        elif item == SmokeGrenade:
            newUse = ThrowSmoke()
        elif item == Flashbang:
            newUse = ThrowFlashbang()
        else:
            newUse.setItem(item)
        newUse.x, newUse.y = self.player.movement.getPosition()
        newUse.direction = self.player.movement.direction
        newUse.power = power
        
    def hit(self, targetPlayer, damage, weapon, headshot = False, bloodAmount = 0):
        if weapon in guns:
            newHit = Hit()
            newHit.setWeapon(weapon)
        else:
            newHit = HeavyHit()
            newHit.setItem(weapon)
        newHit.targetPlayer = targetPlayer.id
        newHit.damage = damage
        newHit.headshot = headshot
        newHit.bloodAmount = bloodAmount
        
    def move(self, direction):
        movement = self.player.movement
        x, y = movement.getPosition()
        movement.set(x, y, direction)
        movement.start()
        newMoving = Moving()
        newMoving.direction = direction
        self.sendType(newMoving)
        
    def changeDirection(self, direction):
        movement = self.player.movement
        x, y = movement.getPosition()
        movement.set(x, y, direction)
        movement.start()
        newChanged = ChangeDirection()
        newChanged.x = x
        newChanged.y = y
        newChanged.direction = direction
        self.sendType(newChanged)
        
    def look(self, direction):
        self.player.look_direction = direction
        newLook = LookDirection()
        newLook.direction = direction
        self.sendType(newLook)
        
    def stop(self):
        movement = self.player.movement
        x, y = movement.getPosition()
        movement.stop()
        newStopped = Stopped()
        newStopped.x = x
        newStopped.y = y
        self.sendType(newStopped)
        
    def sendMessage(self, message):
        newChat = Chat()
        newChat.value = message
        self.sendType(newChat)
        
    def sendTeamMessage(self, message):
        newChat = TeamChat()
        newChat.value = message
        self.sendType(newChat)
        
    def requestSpawn(self, x, y, now = False):
        self._requestTime = time.time()
        self._spawnX = x
        self._spawnY = y
        self.sendType(RequestSpawn())
        
    def _spawnGranted(self):
        self.latency = (time.time()-self._requestTime) * 1000
        self._requestTime = None
        movement = self.player.movement
        movement.set(self._spawnX, self._spawnY, movement.direction)
        self._sendSpawn(self._spawnX, self._spawnY)
        self._spawnX = None
        self._spawnY = None
        self.spawnGranted()
        
    def _sendSpawn(self, x, y):
        newSpawn = Spawn()
        newSpawn.x = x
        newSpawn.y = y
        self.sendType(newSpawn)
        
    def getTimeLeft(self):
        self.sendType(getTimeLeft())
        
    def getNextMap(self):
        self.sendType(getNextMap())
    
class ClientGameFactory(GameFactory, ClientFactory):
    protocol = ClientGameProtocol
    initial = None
    
    def __init__(self, initial):
        """
        @param initial: An InitialMessage object received from
            the server earlier.
        """
        self.initial = initial
        self.key = initial.gameKey
        
    def loadMap(self, mapName):
        """
        Override this method to change the way the client
        loads the map.
        
        @return: L{Map} instance
        """
        import os
        from pySndo.map.loader import Map
        # search path
        path = os.path.join(os.getcwdu(), 'Maps', mapName + '.sdo')
        if not os.path.isfile(path):
            raise Exception('no proper map found')
        newMap = Map(open(path, 'rb').read())
        newMap.name = mapName
        return newMap

class ClientInfoProtocol(InfoProtocol):
    """
    The client info protocol.
    
    @ivar mapReceiver: This is used to keep track of a map
        being received currently.
    @ivar initial: The intitial message received from the server.
    """
    _isClient = True
    initial = None
    
    mapSending = None
    mapReceiver = None
    
    def connectionMade(self):
        self.sendType(GetMapSending())
        self.sendType(GetInitial())
    
    def _gotMapSending(self, mapSending):
        self.mapSending = mapSending.enabled
    
    def packetReceived(self, packet):
        packetType = packet.type
        
        if isinstance(packetType, InitialMessage):
            self.initial = packetType
            self.initialReceived(packetType)
        elif isinstance(packetType, Error):
            self.errorOccured()
        elif isinstance(packetType, MapLine):
            self._mapLineReceived(packetType.value)
        elif isinstance(packetType, MapSending):
            self._gotMapSending(packetType)
            
    # overridables
    
    def initialReceived(self, initial):
        """
        Called after receiving the initial message.
        Here, you would most likely want to join the game
        (see L{joinGame}) or receive the map
        (see L{requestMap}).
        
        @param initial: L{InitialMessage} instance
        """
        
    def errorOccured(self):
        """
        Called when the server has responded with
        an unknown error (map not found, etc)
        """
        
    def mapProgressed(self, currentLines, totalLines):
        """
        Called when progress on a map being received has
        been made.
        
        @param currentLines: Number of lines received in total.
        @param totalLines: Number of lines in total.
        """
        
    def mapReceived(self, map):
        """
        Called when a map has been successfully received.
        
        @param map: L{Map} instance
        """
        
    # actions

    def joinGame(self):
        """
        Call this after receiving the initial message
        to join the actual game
        """
        if not self.initial:
            return False
        newFactory = self.factory.gameFactory(self.initial)
        host = self.transport.getPeer().host
        port = self.transport.getPeer().port
        reactor.connectTCP(host, port-1, newFactory)
        self.transport.loseConnection()
        return newFactory
        
    def requestMap(self, mapName):
        """
        Request a map from the server
        
        @return: False if map sending is not available, True otherwise
        """

        if self.mapSending:
            if self.mapReceiver:
                raise Exception('already receiving a map')
            self.mapReceiver = MapReceiver(mapName)
            self._getNextLine()
            return True
        else:
            return False
            
    def _getNextLine(self):
        newRequest = RequestMapLine()
        newRequest.requestedLine = self.mapReceiver.getNext()
        newRequest.mapName = self.mapReceiver.mapName
        self.sendType(newRequest)
        
    def _mapLineReceived(self, line):
        if not self.mapReceiver:
            raise Exception('received map line, but no map receiver')
        if self.mapReceiver.feed(line):
            # map done
            self.mapReceived(self.mapReceiver.getMap())
            self.mapReceiver = None
        else:
            # do map transfer
            currentLines, totalLines = self.mapReceiver.getProgress()
            self.mapProgressed(currentLines, totalLines)
            self._getNextLine()

class ClientInfoFactory(InfoFactory, ClientFactory):
    """
    The client info factory.
    
    @ivar gameFactory: The factory to use for the game protocol
    """

    protocol = ClientInfoProtocol
    gameFactory = ClientGameFactory