# Copyright (c) 2009 Mathias Kaerlev.
# See LICENSE for details.

# Copyright (c) 2009 Mathias Kaerlev.
# See LICENSE for details.

import time

import random

from random import randrange

import os

import traceback

# twisted

from twisted.internet import reactor, defer

from twisted.internet.task import LoopingCall

from twisted.internet.protocol import ServerFactory

# site connection

from pySndo.network.site.parsers import Server
from pySndo.network.site.transmitters import AddServer, DeleteServer

# network

from makekey import generateKey

from pySndo.exceptions import HackingAttempt

from pySndo.cipher import MMFBlowfish

from pySndo.network.protocol import GameProtocol, GameFactory, \
    InfoProtocol, InfoFactory
    
from pySndo.network.info import GetInitial, RequestMapLine, GetMapSending, \
    InitialMessage, Error, MapLine, MapSending, MapSender
    
from pySndo.network.game import Message

# all packetloaders

from pySndo.network.packetloaders.welcome import Welcome, SetName
from pySndo.network.packetloaders.playerlist import PlayerList
from pySndo.network.packetloaders.map import CurrentMap, SetMap, MapStart, \
    BombedArea, SpawnItem, GotMedipack
from pySndo.network.packetloaders.chat import Chat, TeamChat, Typing, \
    DoneTyping, ServerMessage
from pySndo.network.packetloaders.ping import Ping, Pong
from pySndo.network.packetloaders.join import JoinGame, RequestSpawn, \
    SpawnGranted
from pySndo.network.packetloaders.team import SetTeam, TeamScore, \
    IncrementScore, SelfKill
from pySndo.network.packetloaders.movement import Moving, Stopped, \
    ChangeDirection, LookDirection, Spawn
from pySndo.network.packetloaders.battle import Reload, Shoot, Hit, HeavyHit, \
    Killed, UseItem, ThrowFlashbang, ThrowSmoke, ThrowMolotov, \
    itemToId, weapons, dieWeapons
from pySndo.network.packetloaders.player import PlayerJoined, PlayerLeft, \
    PlayerKicked, PlayerBanned, GotBanned, PlayerKills, SkinChange
from pySndo.network.packetloaders.status import Closing, Restart, ServerFull, \
    TimeLeft, NextMap, ServerStatus

# packetloader bases

from pySndo.network.packetloaders.common import PrefixSend

# misc. network

from pySndo.network.validcheck import validName

from pySndo.idpool import IDPool

from pySndo.network.protection import ConnectionChecker

# logic

from pySndo.game.player import Player

from pySndo.game.broadcast import MessageList

from pySndo.game.weapon import Glock, SIGP, DesertEagle, Benelli, BenelliHit, \
    MP5, Galil, SVD, AK47, Colt, G3A3, SR60, AW50, GasGrenade, HEGrenade, C4, \
    Flashbang, MolotovCocktail, SmokeGrenade, Medipack, Other

from pySndo.game.crate import guns, grenades

MAX_ITEMS_PER_ROUND = 3
MAX_TEXT_LENGTH = 100

_serverPlayer = PlayerList()
_serverPlayer.name = 'Server'
_serverPlayer.latency = 0
_serverPlayer.direction = 16
_serverPlayer.moving = False
_serverPlayer.x = -20
_serverPlayer.y = -20
_serverPlayer.setTeam('Generic')
_serverPlayer.deaths = 0
_serverPlayer.kills = 0
_serverPlayer.ping = 0
_serverPlayer.alive = False
_serverPlayer.typing = False

def _getServerPlayer():
    return _serverPlayer

class ServerGameProtocol(GameProtocol):
    """
    Server game protocol
    
    @ivar messageList: An object keeping track of the 'noticable'
        messages
    @ivar latency: Latency (ping) between the client and server.
    @ivar player: The player representing this connection
    """
    _requestedName = None
    _isClient = False
    messageList = None
    
    latency = None
    _pingCall = None # DelayedCall for the ping check,
                     # canceled upon receiving a response
    _pingTime = None
    _disconnected = False
    
    _itemsUsed = 0 # protection against effing DriftoO >_>
    
    player = None
    
    # implementation-specific
    
    def connectionMade(self):
        self.factory.checker.playerConnected(self)
        self.messageList = MessageList(self._updateMessage)
        reactor.callLater(self.factory.timeOut/1000, self._timedOut)
        
    def connectionLost(self, reason):
        if self._disconnected:
            return
        self._disconnected = True
        self.factory.checker.playerDisconnected(self)
        if self.isAccepted:
            self.factory.connections.remove(self)
            newLeft = PlayerLeft()
            newLeft.id = self.id
            self.factory.broadcastType(newLeft)
            self.factory.idPool.putBackId(self.id)
            if (self.factory.map.data.getGameType() == 'Last Team Standing' 
                    and not self.factory.hasRoundStarted()
                    and self.player.getTeam() != 'Generic'):
                self.factory.endRound()
        
    def _timedOut(self):
        if self._disconnected:
            return
        if not self.isAccepted:
            self.transport.loseConnection()
            self.factory.checker.playerTimedOut(self)
            self.timedOut()
    
    def packetCorrupted(self, realData, e):
        traceback.print_exc()
        self.kick('Server error, %s' % traceback.format_exc())
    
    def packetReceived(self, packet):
        packetType = packet.type
        
        if self._disconnected:
            return
        
        if isinstance(packetType, PrefixSend):
            packetType.id = self.id
        
        # welcome message
        if isinstance(packetType, SetName):
            name = packetType.value
            nameList = [connection.name for connection in self.factory.connections]
            # if server full
            if len(self.factory.connections) == self.factory.maxPlayers:
                self.sendType(ServerFull())
                self.transport.loseConnection()
            elif not validName(name):
                self.kick('Invalid name')
            elif name in nameList:
                self.kick('Name taken')
            elif self.acceptLogin(name):
                self.name = name
                self.isAccepted = True
                newId = self.factory.idPool.popId()
                self.id = newId
                
                newWelcome = Welcome()
                newWelcome.id = newId
                self.sendType(newWelcome)

                if self.factory.map.data.getGameType() in ('Team Deathmatch', 'Last Team Standing'):
                        newScores = TeamScore()
                        newScores.terroristPoints = self.factory.terrorist.score
                        newScores.policePoints = self.factory.police.score
                        self.sendType(newScores)
                    
                newCurrent = CurrentMap()
                newCurrent.value = self.factory.map.name
                self.sendType(newCurrent)
                
                for bombedArea in self.factory.bombedAreas:
                    self.sendType(bombedArea)
                    
                newPlayer = _getServerPlayer()
                self.sendType(newPlayer)
                newPlayer = PlayerList()
                    
                for connection in self.factory.connections:
                    player = connection.player
                    movement = connection.player.movement
                    newPlayer.name = connection.name
                    newPlayer.id = connection.id
                    newPlayer.direction = movement.direction or 24
                    newPlayer.moving = movement.moving or False
                    x, y = movement.getPosition()
                    newPlayer.x = x or -20
                    newPlayer.y = y or -20
                    newPlayer.setTeam(player.getTeam())
                    newPlayer.deaths = player.deaths
                    newPlayer.kills = player.kills
                    newPlayer.latency = int(connection.latency or 0)
                    newPlayer.alive = player.alive
                    newPlayer.typing = player.typing
                    self.sendType(newPlayer)
                    
                newJoin = PlayerJoined()
                newJoin.id = self.id
                newJoin.name = name
                self.factory.broadcastType(newJoin)
                
                # player object
                self.player = Player()
                
                self.factory.connections.append(self)
                
                self.loggedIn(name)
                
        elif isinstance(packetType, JoinGame):
            # The protocol sucks v_v
            return
            
        elif isinstance(packetType, SetTeam):
            if self.player.getTeam() != 'Generic':
                self.player.alive = False
                wasChange = True
                newKill = SelfKill()
                newKill.keepScore = True
                # TODO: not sure.
                self.broadcastType(newKill)
            else:
                wasChange = False

            self.player.setTeam(packetType.getTeam())
            teamObject = self.factory.teamFromName(packetType.getTeam())
            self.broadcastType(packetType)
            
            if wasChange:
                self.teamChanged(teamObject)
            else:
                self.teamJoined(teamObject)
            
            if self.factory.map.data.getGameType() == 'Last Team Standing' \
                    and not self.factory.hasRoundStarted():
                self.factory.endRound()
            
        elif isinstance(packetType, Typing):
            if self.player:
                self.player.typing = True
            self.broadcastType(packetType)

        elif isinstance(packetType, DoneTyping):
            if self.player:
                self.player.typing = False
            self.broadcastType(packetType)

        elif isinstance(packetType, Chat):
            if len(packetType.value) > MAX_TEXT_LENGTH:
                self.kick('Max text length exceeded', True)
                self.factory.connectionFlooding(self.transport.getPeer().host)
                return
            if self.player:
                self.player.typing = False
            if self.acceptMessage(packetType.value):
                self.spoofMessage(packetType.value)
                self.messageReceived(packetType.value)
            else:
                self.broadcastType(DoneTyping())
        
        elif isinstance(packetType, TeamChat):
            self.player.typing = False
            # just so *everyone* knows our guy is done typing
            self.broadcastType(DoneTyping())
            team = self.player.getTeam()
            if team == 'Generic':
                self.kick('Unexpected packet')
                print 'TeamChat but not on team'
            elif self.acceptTeamMessage(packetType.value):
                self.spoofTeamMessage(packetType.value)
                self.teamMessageReceived(packetType.value)
                
        elif isinstance(packetType, PlayerKills):
            # see JoinGame
            # tldr; The protocol sucks v_v
            return
            
        elif isinstance(packetType, SelfKill):
            # see JoinGame
            # tldr; The protocol sucks v_v
            return
        
        elif isinstance(packetType, Pong):
            self._pongReceived()
            
        elif isinstance(packetType, RequestSpawn):
            # can't prevent the player from spawning in LTS mode
            # (but can prevent client from knowing MapStart)
            if self.factory.map.data.getGameType() == 'Last Team Standing':
                return
            if (self.factory.map.data.getGameType() in (
                    'Team Deathmatch')
                    and self.player.getTeam() == 'Generic'):
                print 'Request spawn but not on team?'
                self.kick('Unexpected packet')
                return
            if self.acceptSpawn():
                self.grantSpawn()
            
        elif isinstance(packetType, Spawn):
            # if player tried to spawn (in lts, tdm mode) even though it
            # had not joined a team yet, disconnect it
            if (self.factory.map.data.getGameType() in (
                    'Team Deathmatch', 
                    'Last Team Standing')
                    and self.player.getTeam() == 'Generic'):
                print 'Spawned but not on team?'
                self.kick('Unexpected packet')
                return
            player = self.player
            # reset player
            self._itemsUsed = 0
            player.health = 100
            player.alive = True
            player.movement.stop()

            player.movement.set(packetType.x, packetType.y, 24)
            packetType.latency = int(self.latency or 0)
            self.broadcastType(packetType)
            self.playerSpawned(packetType.x, packetType.y)
            
        elif isinstance(packetType, Moving):
            if not self._aliveCheck():
                return
            movement = self.player.movement
            x, y = movement.getPosition()
            movement.set(x, y, packetType.direction)
            movement.start()
            self.broadcastType(packetType)
            self.playerMoving()
            
        elif isinstance(packetType, Stopped):
            if not self._aliveCheck():
                return
            movement = self.player.movement
            movement.stop()
            # TODO: add checks
            movement.set(packetType.x, packetType.y, movement.direction)
            self.broadcastType(packetType)
            self.playerStopped()
            
        elif isinstance(packetType, ChangeDirection):
            if not self._aliveCheck():
                return
            movement = self.player.movement
            movement.stop()
            # TODO: add checks
            movement.set(packetType.x, packetType.y, packetType.direction)
            movement.start()
            self.broadcastType(packetType)
            self.changedDirection()

        elif isinstance(packetType, LookDirection):
            if not self._aliveCheck():
                return
            if self.acceptLook(packetType.direction):
                self.player.look_direction = packetType.direction
                self.look(packetType.direction)
                self.playerLooked(packetType.direction)
            
        elif isinstance(packetType, Shoot):
            # TODO: add checks
            arguments = (packetType.getWeapon(), packetType.direction, packetType.showOnly,
                         packetType.hitX, packetType.hitY)
            if self.acceptShot(*arguments):
                self.shoot(*arguments)
                self.playerShot(*arguments)
            
        elif isinstance(packetType, UseItem):
            # TODO: add checks
            if self._checkUsed():
                return
            item = packetType.getItem()
            x, y = self.player.movement.getPosition()
            """
            packetType.x, packetType.y = x, y
            """
            self.broadcastType(packetType)
            self.usedItem(item, packetType.direction, packetType.power)
            if item == C4:
                self.factory._bombCalls.append(
                    reactor.callLater(packetType.power, self._addBombedArea, x, y)
                )
            
        elif packetType.__class__ in (ThrowMolotov, ThrowFlashbang, ThrowSmoke):
            # TODO: add checks
            if self._checkUsed():
                return
            """
            x, y = self.player.movement.getPosition()
            packetType.x, packetType.y = x, y
            """
            self.broadcastType(packetType)
            self.usedItem(packetType.getWeapon(), packetType.direction, packetType.power)
            
        elif isinstance(packetType, Reload):
            if not self._aliveCheck():
                return
            # TODO: add checks
            x, y = self.player.movement.getPosition()
            packetType.x, packetType.y = x, y
            self.player.reloading = packetType.reloading
            self.broadcastType(packetType)
            self.playerReload(packetType.reloading)
            
        elif packetType.__class__ in (Hit, HeavyHit):
            if self.factory._roundStarting:
                return
            try:
                targetPlayer = self.factory.playerFromId(packetType.targetPlayer)
            except IndexError:
                return
            if not targetPlayer.player.alive:
                return
            if not self.factory.friendlyFire and targetPlayer != self and (
                self.player.getTeam() == targetPlayer.player.getTeam()):
                return
            if not self.player.alive and isinstance(packetType, Hit):
                return
            if isinstance(packetType, Hit):
                if not self.player.alive:
                    return
                weapon = packetType.getWeapon()
            else:
                weapon = packetType.getItem()
            arguments = (targetPlayer, packetType.damage, weapon,
                         packetType.headshot, packetType.bloodAmount)
            if self.acceptHit(*arguments):
                killed = self.hit(*arguments)
                self.playerHit(*arguments)
                if killed:
                    self.playerKilled(targetPlayer, dieWeapons[weapon])
                    
        elif isinstance(packetType, SpawnItem):
            newCrate = packetType.getCrateType()(self.factory.crates)
            self.factory.crates.append(newCrate)
            newCrate.x = packetType.x
            newCrate.y = packetType.y
            for key, value in self.factory.weapons.iteritems():
                if value == False and key in newCrate.possibleContent:
                    newCrate.possibleContent.remove(key)
            newCrate.removeCrate().addCallback(self.crateDestroyed)
            self.crateSpawned(newCrate)
            
        elif isinstance(packetType, GotMedipack):
            if not self._aliveCheck():
                return
            self.player.health += Medipack.heal
                    
        elif isinstance(packetType, PlayerBanned):
            # see JoinGame
            # tldr; The protocol sucks v_v
            return
            
        elif isinstance(packetType, TimeLeft):
            self.timeLeftReceived()

        elif isinstance(packetType, NextMap):
            self.nextMapReceived()
            
        else:
            print 'unknown packet'
            raise ValueError('unknown packet: %s %s %s %s' % (packet.way, packet.id, 
                packetType.__class__.__name__, vars(packetType)))

    def _aliveCheck(self, kick = False):
        if not self.player.alive:
            if kick:
                self.kick('Unexpected packet')
            return False
        return True
    
    def _checkUsed(self):
        if self._itemsUsed > MAX_ITEMS_PER_ROUND:
            self.kick('Max items per round exceeded', True)
            self.factory.connectionFlooding(self.transport.getPeer().host)
            return True
        self._itemsUsed += 1
        return False
        
    def _addBombedArea(self, x, y):
        newBombed = BombedArea()
        newBombed.x = x
        newBombed.y = y
        self.factory.bombedAreas.append(newBombed)

    def sendType(self, packetType):
        newMessage = Message()
        newMessage.type = packetType
        if isinstance(packetType, PrefixSend) and not packetType.id:
            packetType.id = self.factory._serverId # server id
        self.sendLine(newMessage.generate())

    def broadcastType(self, packetType):
        """
        Broadcast a packetloader to all clients but
        this one.
        """
        if isinstance(packetType, PrefixSend) and not packetType.id:
            packetType.id = self.id # ourselves
        self.factory.broadcastType(packetType, [self])
        
    def _updateMessage(self, messageList):
        newMessage = ServerMessage()
        newMessage.value = '\n'.join(messageList.getMessages())
        self.sendType(newMessage)

    # overridables
    
    def timedOut(self):
        """
        Called when the client has timed out the connection and
        has been disconnected. It might be a good idea to put
        the client in a ban list, as it is most likely a crasher.
        """
    
    def loggedIn(self, name):
        """
        Called upon logging in.
        
        @param name: The name of the client, None if
        protocol is client.
        """
        
    def messageReceived(self, message):
        """
        Called when the player has sent a chat message.
        """
        
    def teamMessageReceived(self, message):
        """
        (LTS and TDM mode only)
        Called when the player has sent a team message.
        """
        
    def teamJoined(self, team):
        """
        (LTS and TDM mode only)
        Called upon joining a team.
        
        @type team: L{Police} or L{Terrorist} instance.
        """

    def teamChanged(self, team):
        """
        (LTS and TDM mode only)
        Called upon changing current team.
        
        @type team: L{Police} or L{Terrorist} instance.
        """
        
    def playerSpawned(self, x, y):
        """
        Called when a player has spawned.
        """
        
    def playerMoving(self):
        """
        Called when the player has started movement.
        """
        
    def playerLooked(self, lookDirection):
        """
        Called when a player has switched the direction
        it is aiming at.
        """
        
    def changedDirection(self):
        """
        Called when the player changed direction while moving.
        """
        
    def playerStopped(self):
        """
        Called when the player has stopped movement.
        
        @type player: L{OtherClient} instance
        """

    def usedItem(self, item, direction, power):
        """
        Called when the player has used a special item.

        @type item: See L{pySndo.game.weapon.weaponList} for possible values.
        """
        
    def playerReload(self, reloading):
        """
        Called when the player has started/stopped 
        reloading.
        """

    def playerShot(self, weapon, direction, showOnly, hitX, hitY):
        """
        Called when the player has shot without hitting.
        
        @type weapon: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param showOnly: If True, hitX and hitY are not specified
        
        @param direction: The current look direction of the player
        
        @param hitX: The X coordinate of the obstacle the bullet has
        collided with.
        @param hitY: The Y coordinate of the obstacle the bullet has
        collided with.
        """
        
    def playerHit(self, targetPlayer, damage, weapon, headshot = False, bloodAmount = 0):
        """
        Called upon receiving a 'hit' from the client.
        
        @param player: The player being hit
        @type player: L{ServerGameProtocol} instance
        
        @type weapon: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param damage: Damage the target player is being hit with
        """
        
    def playerKilled(self, player, dieType):
        """
        Called when the player has killed another player.
        
        @param player: The killed player
        @type player: L{ServerGameProtocol} instance
        @param dieType: See L{dieWeapons}
        @type dieType: str
        """
    
    def timeLeftReceived(self):
        """
        Called when a player has asked for time left
        on map.
        """
        
    def nextMapReceived(self):
        """
        Called when a player has asked what the
        next map in the map cycle is.
        """
        
    # accept-like overridables
    
    def acceptLogin(self, name):
        """
        Return False to prevent from accepting
        the client (useful for ban implementation)
        
        @param name: The requested name.
        """
        return True
    
    def acceptMessage(self, message):
        """
        Return False to block the sending of this message
        to the other players.
        """
        return True

    def acceptTeamMessage(self, message):
        """
        (LTS and TDM mode only)
        Return False to block the sending of this team message
        to the other players.
        """
        return True

    def acceptMapStart(self):
        """
        (LTS mode only)
        Return False to effectively prevent the player from spawning
        in LTS mode.
        """
        return True
        
    def acceptSpawn(self):
        """
        (DM and TDM mode only)
        Return False to block the server from sending
        a granted spawn to the server
        (see L{grantSpawn})
        """
        return True
        
    def acceptLook(self, lookDirection):
        """
        Return False to prevent the new look direction
        from being sent to the clients.
        """
        return True
        
    def acceptShot(self, weapon, direction, showOnly, hitX, hitY):
        """
        Return False to block the missed player shot to be sent
        to the other players.
        
        @type weapon: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param direction: The current look direction of the player
        
        @param showOnly: If True, hitX and hitY are not specified
        
        @param hitX: The X coordinate of the obstacle the bullet has
        collided with.
        @param hitY: The Y coordinate of the obstacle the bullet has
        collided with.
        """
        return True
        
    def acceptHit(self, targetPlayer, damage, weapon, headshot = False, bloodAmount = 0):
        """
        Return False to block the player hit from being applied.
        
        @param player: The player being hit
        @type player: L{ServerGameProtocol} instance
        
        @type weapon: See L{pySndo.game.weapon.weaponList} for possible values.
        
        @param damage: Damage the target player is being hit with
        """
        return True
    
    def acceptTimeLeft(self):
        """
        Return False to ignore time left request.
        """
        return True
        
    def acceptNextMap(self):
        """
        Return False to ignore next map request.
        """
        return True
        
    # action
    
    def ping(self, maxPing = None):
        """
        Ping this client, and disconnect if pong timed out.
        
        @param maxPing: Max number of miliseconds this connection
            has to reply to the ping (defaults to factory.maxPing).
        """
        if self._pingCall and self._pingCall.active():
            # still waiting for a ping reply, so return
            return False
        maxPing = maxPing or self.factory.maxPing
        self._pingCall = reactor.callLater(maxPing*2/1000.0, self._pingTimedOut)
        self._pingTime = time.time()
        newPing = Ping()
        self.sendType(newPing)
        return True
        
    def _pingTimedOut(self):
        if self._disconnected:
            return
        maxPing = int((time.time() - self._pingTime) / 2 * 1000)
        self.kick('Max ping exceeded (%s)' % (maxPing))
        
    def _pongReceived(self):
        """
        Called upon receiving a pong
        """
        # client timed out?
        if not self._pingCall.active():
            return
        self._pingCall.cancel()
        self.latency = ((time.time() - self._pingTime) * 1000) / 2
        self._pingTime = None
    
    def sendStatus(self, message):
        """
        Send a server status message (to this client only).
        """
        newStatus = ServerStatus()
        for line in message.split('\n'):
            newStatus.message = line
            self.sendType(newStatus)
        
    def sendMessage(self, *arg, **kw):
        """
        Send a noticable message (to this client only).
        """
        self.messageList.addMessage(*arg, **kw)
        
    def sendMessages(self, *arg, **kw):
        """
        Send a noticable message (to this client only).
        """
        self.messageList.addMessages(*arg, **kw)
        
    def kick(self, reason = None, ban = False):
        """
        Kick/ban a player from the server
        
        @param reason: A string reason to specify the reason
            for the kick.
        @param ban: Tell the client it was banned from the server
            (you need to keep a list of banned clients yourself)
        """
        if ban:
            value = 'You have been banned'
            broadcast = '%s has been banned' % self.name
            newLeave = PlayerBanned()
        else:
            value = 'You have been kicked'
            broadcast = '%s has been kicked' % self.name
            newLeave = PlayerKicked()
        if reason:
            value += ', reason: %s' % reason
            broadcast += ', reason: %s' % reason
        if not self.isAccepted:
            newWelcome = Welcome()
            newWelcome.id = 2
            self.sendType(newWelcome)
            self.sendType(_getServerPlayer())

        self.sendMessage(value)
        if self.isAccepted:
            newLeave.playerId = self.id
            self.factory.broadcastType(newLeave)
            self.factory.broadcastMessage(broadcast)
            if ban:
                self.sendType(GotBanned())
        else:
            newLeave.playerId = 2
            self.sendType(newLeave)

        self.connectionLost(None)
        reactor.callLater(5, self.transport.loseConnection)
        
    # spoof client-like actions
    
    def shoot(self, weapon, direction, showOnly = True, hitX = None, hitY = None):
        newShoot = Shoot()
        newShoot.direction = direction
        newShoot.showOnly = showOnly
        newShoot.hitX = hitX
        newShoot.hitY = hitY
        newShoot.setWeapon(weapon)
        self.broadcastType(newShoot)
        
    def hit(self, targetPlayer, damage, weapon, headshot = False, bloodAmount = 0):
        if weapon in itemToId:
            newHit = HeavyHit()
            newHit.setItem(weapon)
        elif weapon in weapons.values():
            newHit = Hit()
            newHit.setWeapon(weapon)
        else:
            raise ValueError('unknown weapon specified')
        newHit.id = self.id
        newHit.targetPlayer = targetPlayer.id
        newHit.damage = damage
        newHit.headshot = headshot
        newHit.bloodAmount = bloodAmount
        self.broadcastType(newHit)
        if weapon == AW50:
            damage = 100 # kill
        elif weapon == SR60 and headshot == True:
            damage = 100 # kill
        if targetPlayer.player.damage(damage):
            self.kill(targetPlayer, dieWeapons[weapon])
            return True
        else:
            return False
            
    def kill(self, targetPlayer, dieType = 'Shot'):
        gameType = self.factory.map.data.getGameType()

        player = targetPlayer.player
        player.alive = False
        player.deaths += 1
        player.movement.stop()

        # if player is from same team
        playerTeam = self.player.getTeam()
        
        teamObject = self.factory.teamFromName(playerTeam)
        
        if gameType != 'Deathmatch' and playerTeam == player.getTeam():
            self.player.kills -= 2
        else:
            self.player.kills += 1
            if gameType == 'Team Deathmatch':
                teamObject.score += 1

        newKilled = Killed()
        newKilled.killingPlayer = self.id
        newKilled.killedPlayer = targetPlayer.id
        newKilled.setDieType(dieType)
        self.factory.broadcastType(newKilled)
        
        newKills = PlayerKills()
        newKills.id = self.id
        newKills.kills = self.player.kills
        self.broadcastType(newKills)
        
        if gameType == 'Last Team Standing':
            if not self.factory.hasRoundStarted():
                self.factory.endRound()
        
    def look(self, direction):
        newLook = LookDirection()
        newLook.direction = direction
        self.broadcastType(newLook)
        
    def spoofMessage(self, message):
        newMessage = Chat()
        newMessage.value = message
        self.broadcastType(newMessage)
        
    def spoofTeamMessage(self, message):
        newMessage = TeamChat()
        newMessage.value = message
        newMessage.id = self.id
        teamObject = self.factory.teamFromName(self.player.getTeam())
        teamConnections = teamObject.getConnections()
        for connection in teamConnections:
            connection.sendType(newMessage)
        
    def grantSpawn(self):
        self.sendType(SpawnGranted())

class ServerGameFactory(GameFactory, ServerFactory):
    """
    The server game factory.
    
    @ivar maxPlayers: Max amount of players allowed on this server
    @ivar friendlyFire: Whether teammates can hit each other
    @ivar key: Key used to encrypt the connection.
    @ivar idPool: Pool of IDs used to assign an unique ID
        to every connection.
    @ivar round: L{Round} object used to keep track of rounds.
    @ivar roundStartDelay: Number of seconds to delay the starting
        of the round.
    @ivar weapons: Weapons available on this server
    """
    protocol = ServerGameProtocol
    key = None
    idPool = None
    _serverId = None
    
    maxPlayers = 50
    
    friendlyFire = True
    
    pinger = None
    pingInterval = 8 # in seconds
    maxPing = 300 # in ms
    timeOut = 3000 # connection timeout in ms
    
    round = None
    roundStartDelay = 4 # default round start delay in seconds
    _roundStarting = False
    
    weapons = {
        DesertEagle : True,
        Benelli : True,
        MP5 : True,
        AK47 : True,
        SVD : True,
        Colt : True,
        G3A3 : True,
        AW50 : True,
        Galil : True,
        SR60 : True
    }
    
    checker = None
    
    _bombCalls = None
    
    def __init__(self, key = None):
        """
        @param key: If specified, this will be the key used 
            for the game connection (if not specified, it is
            generated automatically)
        """
        key = key or generateKey(20)
        self.key = key
        checker = ConnectionChecker()
        checker.addHandler(self._banConnection)
        self.checker = checker
        
    def _banConnection(self, ip):
        for connection in [connection for connection in self.connections
                            if connnection.getPeer().host == ip]:
            connection.transport.loseConnection()
        self.connectionFlooding(ip)
    
    def startFactory(self):
        GameFactory.startFactory(self)
        self._bombCalls = []
        self.cipher = MMFBlowfish(self.key)
        self.idPool = IDPool(2)
        self._serverId = 1
        pinger = LoopingCall(self.pingPlayers)
        pinger.start(self.pingInterval, False)
        self.pinger = pinger
        
        # for shutdown
        reactor.addSystemEventTrigger('before', 'shutdown', self.shutdownCalled)
    
    # overridables
    
    def connectionFlooding(self, ip):
        """
        Called when a connection has been detected as
        a so called DOSer (or flooder), and all clients from this IP has
        been disconnected. You're advised to ban this client forever.
        """

    def mapChanged(self, newMap):
        """
        Called upon map change.
        
        @type newMap: L{Map}
        """

    def roundEnded(self, winningTeam):
        """
        (LTS mode only)
        Called upon the end of the round
        
        @param winningTeam: The winner of this round
        @type winningTeam: L{Terrorist} or L{Police}
        """

    def acceptRoundStart(self):
        """
        Return False to block the server
        from starting the round
        """
        return True
        
    def roundStarted(self):
        """
        (LTS mode only)
    
        Called upon round start
        """
        
    # actions
    
    def changeMap(self, map):
        self.map = map
            
        setMap = SetMap()
        setMap.mapName = map.name
        setMap.mapCrc = map.computeChecksum()
        self.broadcastType(setMap)
        
        self.bombedAreas = []
        
        for connection in self.connections:
            connection.player.alive = False
            connection.player.setTeam('Generic')
            connection.player.movement.stop()
            connection.player.deaths = 0
            connection.player.kills = 0
        
        if (not self.hasRoundStarted() and 
                map.data.getGameType()) == 'Last Team Standing':
            self.startRound()
    
    def startRound(self, delay = None):
        if self.acceptRoundStart():
            reactor.callLater(delay or self.roundStartDelay, self._roundStart)
            self._roundStarting = True

    def _roundStart(self):
        for bombCall in self._bombCalls:
            if bombCall.active():
                bombCall.cancel()
        self._bombCalls = []
        self.bombedAreas = []
        self._roundStarting = False
        newMapStart = MapStart()
        for connection in self.connections:
            if connection.acceptMapStart():
                connection.sendType(newMapStart)
                if (self.map.data.getGameType() == 'Last Man Standing'
                    and connection.player.getTeam() == 'Generic'):
                    continue
                #connection.player.alive = True # fix?

    def endRound(self, newRoundDelay = None):
        terrorist, police = self.aliveTeamNumbers()
        if terrorist > police:
            team = self.terrorist
        elif terrorist < police:
            team = self.police
        else:
            self.broadcastMessage('Round ended in a draw!')
            self.startRound(newRoundDelay)
            return
        team.score += 1
        newIncrement = IncrementScore()
        newIncrement.setTeam(team.getName())
        self.broadcastType(newIncrement)
        self.startRound(newRoundDelay)
                
        
    def hasRoundStarted(self):
        terrorist, police = self.aliveTeamNumbers()
        #print 'alive team numbers: %s %s' % (terrorist, police)
        return (terrorist > 0 and police > 0
                or self._roundStarting)
    
    def broadcastMessage(self, *arg, **kw):
        """
        Broadcast a noticable message to all clients.
        """
        for connection in self.connections:
            connection.sendMessage(*arg, **kw)
        
    def broadcastStatus(self, message):
        """
        Broadcast a server status to all clients.
        """
        for connection in self.connections:
            connection.sendStatus(message)
        
    def broadcastType(self, packetType, notClients = []):
        """
        Broadcast a packetloader to all clients.
        
        @param notClients: List of clients not to send
            this packet to.
        """
        # set the id to the server if it hasn't been specified yet
        if isinstance(packetType, PrefixSend) and not packetType.id:
            packetType.id = self._serverId

        for connection in [connection for connection in 
                           self.connections if 
                           not connection in notClients]:
            connection.sendType(packetType)
    
    def broadcastClose(self):
        self.broadcastType(Closing())
        
    def pingPlayers(self):
        """
        Ping all players and record latency.
        """
        for connection in self.connections:
            connection.ping()
            
    def shutdownCalled(self):
        self.broadcastClose()
        d = defer.Deferred()
        reactor.callLater(3, d.callback, None)
        return d
        
class ServerInfoProtocol(InfoProtocol, ServerFactory):
    """
    Server info protocol
    
    (note: it might be a better idea to disconnect/kick the server
    at the game server, as you can provide a reason for the kick
    that way)
    
    @ivar mapSender: A L{MapSender} object that keeps track
        of the map being sent
    """
    
    _isClient = False
    
    _timedOutCall = None
    
    mapSender = None
    
    def connectionMade(self):
        self.factory.connections.append(self)
        self.factory.gameFactory.checker.playerConnected(self)
        self._timedOutCall = reactor.callLater(self.factory.timeOut, 
                                               self._timedOut)

    def connectionLost(self, reason):
        self.factory.connections.remove(self)
        self.factory.gameFactory.checker.playerDisconnected(self)
        if self._timedOutCall.active():
            self._timedOutCall.cancel()
    
    def _timedOut(self):
        self.transport.loseConnection()
        self.factory.gameFactory.checker.playerTimedOut(self)
        self.timedOut()
        
    def packetCorrupted(self, realData, e):
        traceback.print_exc()
        self.transport.loseConnection()
    
    def packetReceived(self, packet):
        """
        InitialMessage, Error, MapLine, MapSending
        
        GetInitial, RequestMapLine, GetMapSending
        """
        packetType = packet.type
        
        if self._timedOutCall.active():
            self._timedOutCall.cancel() # cancel timeout, client is valid (we got data)
        
        if isinstance(packetType, GetInitial):
            gameFactory = self.factory.gameFactory
            map = gameFactory.map
            connections = gameFactory.connections
            if self.acceptInitial():
                newInitial = InitialMessage()
                newInitial.loadFromMap(map)
                newInitial.welcome = self.factory.welcome
                newInitial.currentPlayers = len(connections)
                newInitial.maxPlayers = gameFactory.maxPlayers
                newInitial.gameKey = gameFactory.key
                newInitial.weapons = gameFactory.weapons
                newInitial.password = self.factory.password
                self.sendType(newInitial)
                self.initialSent()
            
        elif isinstance(packetType, RequestMapLine):
            if self._timedOutCall.active():
                self.transport.loseConnection()
                return
            map = self.factory.gameFactory.map
            # requested map line, but map sending not enabled
            if not self.factory.mapSending:
                self.transport.loseConnection()
                return
            # if current map differs from requested map
            if packetType.mapName != map.name:
                self.transport.loseConnection()
                return
            # if a map sender has not been created yet
            if not self.mapSender:
                # 0 is start, so valid
                if packetType.requestedLine == 0:
                    # just exit if usercode didn't accept
                    if not self.acceptMapTransfer(map):
                        return
                    self.mapSender = MapSender(map)
                    self.mapTransferStarted(map)
                # was something else, can't start from inside the map,
                # so must be hacking attempt
                else:
                    self.transport.loseConnection()
                    return
            # if we actually have a map sender
            if self.mapSender:
                newLine = MapLine()
                try:
                    mapLine, isLast = self.mapSender.getLine(packetType.requestedLine)
                # if client requested something else than
                # expected, disconnect
                except HackingAttempt:
                    self.transport.loseConnection()
                    return
                newLine.value = mapLine
                self.sendType(newLine)
                if isLast:
                    # was last, so we destroy the mapsender
                    self.mapSender = None
                    self.mapSent(map)
                else:
                    currentLines, totalLines = self.mapSender.getProgress()
                    self.mapProgress(map, currentLines, totalLines)
        
        elif isinstance(packetType, GetMapSending):
            if self.acceptMapSending():
                self.sendTransferEnabled(self.factory.mapSending)
                self.mapTransferringSent()
    # overridables
    
    def timedOut(self):
        """
        Called when the client has timed out the connection and
        has been disconnected. It might be a good idea to put
        the client in a ban list, as it is most likely a crasher.
        """
    
    def initialSent(self):
        """
        Called when the initial message
        has been sent to the client.
        """
        
    def mapTransferStarted(self, map):
        """
        Called when a map transfer has started.
        """
        
    def mapProgress(self, map, currentLines, totalLines):
        """
        Called upon progress on a map transfer
        has been made.
        
        @type map: L{Map} object
        @param currentLines: Lines transferred to this point
        @param totalLines: Lines in total
        """
        
    def mapSent(self, map):
        """
        Called when a map transfer has ended.
        
        @type map: L{Map object}
        """
    
    def mapTransferringSent(self):
        """
        Called when map sending has been
        sent to the client.
        """
    
    # accept-like overridables
    
    def acceptInitial(self):
        """
        Return False to block the server
        from sending the initial message.
        (not sure what this could be useful for, though)
        """
        return True
    
    def acceptMapSending(self):
        """
        Return False to block the server
        from telling the client whether
        map sending is enabled.
        """
        return True
    
    def acceptMapTransfer(self, map):
        """
        Return False to block the server from
        sending the map
        (you need to call L{sendError} to tell the
        client the map could not be sent yourself)
        """
        return True
        
    # actions
    
    def sendError(self):
        """
        Tells the client an error occured during
        map transfer
        """
        self.sendType(Error())
        
    def sendTransferEnabled(self, enabled):
        """
        Tells the client whether map transferring is
        enabled.
        """
        newSending = MapSending()
        newSending.enabled = enabled
        self.sendType(newSending)

class ServerInfoFactory(InfoFactory, ServerFactory):
    """
    @ivar timeOut: Connection timeout in ms
    @ivar gameFactory: The game server this info server
        is connected to
    @ivar mapSending: Set to True if map sending should be enabled,
        False if otherwise
    @ivar password: Password for this server (None if no password).
    @ivar welcome: Welcome message for this server
    @ivar maxPlayers: The max amount of players allowed to
        play on the server
    """
    protocol = ServerInfoProtocol
    
    connections = None

    timeOut = 3000 # timeout in ms
    gameFactory = None
    welcome = ''
    mapSending = True
    password = None
    
    def __init__(self, gameFactory):
        self.gameFactory = gameFactory
        
    def startFactory(self):
        InfoFactory.startFactory(self)
        self.gameFactory.checker.addHandler(self._banConnection)
        self.connections = []
        
    def _banConnection(self, ip):
        for connection in [connection for connection in self.connections
                            if connnection.getPeer().host == ip]:
            connection.transport.loseConnection()
        
    def connectionFlooding(self, ip):
        """
        Called when a connection has been detected as
        a so called DOSer (or flooder), and all clients from this IP has
        been disconnected. You're advised to ban this client forever.
        """
